/*
    Valtra Tractors
    Copyright 2004 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdio.h>
#include <math.h>
#include "spline.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif


class SplinePoint {
public:
	float x, y, relx, rely;
	SplinePoint(float a, float b, float c, float d) : x(a), y(b), relx(c), rely(d) {}
	SplinePoint() {}
};

class SplineCoords {
public:
	SplinePoint points[3];
	int usedPoints;
	SplineCoords() {
		usedPoints = 0;
	}
	void pushPoint(const SplinePoint &sp) {
		points[usedPoints++] = sp;
	}
};

class TileSplines {
public:
	SplineCoords coords[3];
	int usedCoords;
	TileSplines() {
		usedCoords = 0;
	}
	void pushCoords(const SplineCoords &sc) {
		coords[usedCoords++] = sc;
	}
};


static TileSplines tileSplines[10];
static bool tileSplinesInited = false;
static int tileWidth, tileHeight;

void initSplines(int tileW, int tileH) {
	if (tileSplinesInited) return;
	tileSplinesInited = true;

	tileWidth = tileW;
	tileHeight = tileH;

	SplineCoords sc;
	sc.pushPoint(SplinePoint(37,66,0,0));
	sc.pushPoint(SplinePoint(37,0,0,0));
	tileSplines[0].pushCoords(sc);

	sc = SplineCoords();
	sc.pushPoint(SplinePoint(37,66,0,-20));
	sc.pushPoint(SplinePoint(10,16,-17,-10));
	tileSplines[1].pushCoords(sc);

	sc = SplineCoords();
	sc.pushPoint(SplinePoint(37,0,0,15));
	sc.pushPoint(SplinePoint(10,16,-17*0.8,-10*0.8));
	tileSplines[2].pushCoords(sc);

	sc = SplineCoords();
	sc.pushPoint(SplinePoint(10,16,0,0));
	sc.pushPoint(SplinePoint(66,49,0,0));
	tileSplines[3].pushCoords(sc);
	sc = SplineCoords();
	sc.pushPoint(SplinePoint(9,48,0,0));
	sc.pushPoint(SplinePoint(66,16,0,0));
	tileSplines[3].pushCoords(sc);

	sc = SplineCoords();
	sc.pushPoint(SplinePoint(37,66,0,-20));
	sc.pushPoint(SplinePoint(10,16,-17,-10));
	tileSplines[4].pushCoords(sc);
	sc = SplineCoords();
	sc.pushPoint(SplinePoint(37,0,0,20));
	sc.pushPoint(SplinePoint(9,48,-17,10));
	tileSplines[4].pushCoords(sc);

	sc = SplineCoords();
	sc.pushPoint(SplinePoint(37,66,0,0));
	sc.pushPoint(SplinePoint(37,0,0,0));
	tileSplines[5].pushCoords(sc);
	sc = SplineCoords();
	sc.pushPoint(SplinePoint(10,16,17,10));
	sc.pushPoint(SplinePoint(66,16,17,-10));
	tileSplines[5].pushCoords(sc);
	sc = SplineCoords();
	sc.pushPoint(SplinePoint(9,48,17,-10));
	sc.pushPoint(SplinePoint(66,48,17,10));
	tileSplines[5].pushCoords(sc);

	sc = SplineCoords();
	sc.pushPoint(SplinePoint(37,66,0,0));
	sc.pushPoint(SplinePoint(37,0,0,0));
	tileSplines[6].pushCoords(sc);
	sc = SplineCoords();
	sc.pushPoint(SplinePoint(10,16,17,10));
	sc.pushPoint(SplinePoint(66,16,17,-10));
	tileSplines[6].pushCoords(sc);

	sc = SplineCoords();
	sc.pushPoint(SplinePoint(37,66,0,0));
	sc.pushPoint(SplinePoint(37,0,0,0));
	tileSplines[7].pushCoords(sc);
	sc = SplineCoords();
	sc.pushPoint(SplinePoint(10,16,17*0.8,10*0.8));
	sc.pushPoint(SplinePoint(51,33,0,20));
	sc.pushPoint(SplinePoint(9,48,-17*0.8,10*0.8));
	tileSplines[7].pushCoords(sc);

	sc = SplineCoords();
	sc.pushPoint(SplinePoint(37,66,0,0));
	sc.pushPoint(SplinePoint(37,0,0,0));
	tileSplines[8].pushCoords(sc);
	sc = SplineCoords();
	sc.pushPoint(SplinePoint(10,16,0,0));
	sc.pushPoint(SplinePoint(66,49,0,0));
	tileSplines[8].pushCoords(sc);
	sc = SplineCoords();
	sc.pushPoint(SplinePoint(9,48,0,0));
	sc.pushPoint(SplinePoint(66,16,0,0));
	tileSplines[8].pushCoords(sc);

	sc = SplineCoords();
	sc.pushPoint(SplinePoint(37,66,0,-20));
	sc.pushPoint(SplinePoint(10,16,-17,-10));
	tileSplines[9].pushCoords(sc);
	sc = SplineCoords();
	sc.pushPoint(SplinePoint(9,48,17,-10));
	sc.pushPoint(SplinePoint(66,48,17,10));
	tileSplines[9].pushCoords(sc);
	sc = SplineCoords();
	sc.pushPoint(SplinePoint(37,0,0,15));
	sc.pushPoint(SplinePoint(66,16,17*0.8,-10*0.8));
	tileSplines[9].pushCoords(sc);
}





Spline::Spline(const FPoint &p1, const FPoint &relpoint1, const FPoint &p2, const FPoint &relpoint2) {
	start = p1;
	end = p2;

	waypoint1 = p1 + relpoint1;
	waypoint3 = p2 + relpoint2;
	waypoint2 = (waypoint1+waypoint3)/2;

	float l1 = getLength(start,waypoint1,waypoint2,&len1,&len2);
	float l2 = getLength(waypoint2,waypoint3,end,&len3,&len4);
	if (l1+l2==0) divpos = 1;
	else divpos = l1/(l1+l2);
	totlen = l1+l2;
}

Spline::~Spline() {
}

float Spline::getLength(const FPoint &p1, const FPoint &p2, const FPoint &p3, float *lp1, float *lp2) {
	FPoint tempP1 = p2-p1;
	FPoint tempP2 = p3-p2;
	float l1 = tempP1.abs();
	float l2 = tempP2.abs();
	*lp1 = l1;
	*lp2 = l2;
	if (l2>l1) {
		float tmp = l2;
		l2 = l1;
		l1 = tmp;
	}
	float dot = FPoint::dot(tempP1,tempP2);
	if (l1>0)
		dot /= l1;
	if (l2>0)
		dot /= l2;
	return l1 + l2*(0.5+0.5*dot);
}

float Spline::getLength() {
	return totlen;
}

void Spline::calc(float pos, FPoint *out) {
	if (pos>1) pos = 1;
	if (pos<0) pos = 0;
	FPoint *p1 = &start;
	FPoint *p2 = &waypoint1;
	FPoint *p3 = &waypoint2;
	float l1 = len1;
	float l2 = len2;
	if (pos<divpos) {
		pos = pos/divpos;
	} else {
		p1 = &waypoint2;
		p2 = &waypoint3;
		p3 = &end;
		l1 = len3;
		l2 = len4;
		pos = (pos-divpos)/(1-divpos);
	}
	if (l1!=l2)
		pos = (l1-sqrt(l1*l1+(l2-l1)*((l1+l2)*pos)))/(l1-l2);
	FPoint tempP1 = (*p1)*(1-2*pos+pos*pos);
	FPoint tempP2 = (*p2)*(2*pos-2*pos*pos);
	FPoint tempP3 = (*p3)*(pos*pos);
	tempP1 = tempP1 + tempP2 + tempP3;
	*out = tempP1;
}


CompSpline::CompSpline() {
	inited = false;
	totLen = 0;
}

CompSpline::~CompSpline() {
	for (unsigned int i = 0; i < splines.size(); i++)
		delete splines[i];
}

void CompSpline::pushPoint(float x, float y, float relx, float rely) {
	if (!inited) {
		inited = true;
		lastPos = FPoint(x,y);
		lastRel = FPoint(relx,rely);
		return;
	}
	FPoint newPos = FPoint(x,y);
	FPoint newRel = FPoint(-relx,-rely);
	Spline *spline = new Spline(lastPos,lastRel,newPos,newRel);
	float length = spline->getLength();
	totLen += length;
	lengths.push_back(length);
	splines.push_back(spline);
	lastPos = newPos;
	lastRel = -newRel;
}

void CompSpline::calc(float pos, FPoint *out) {
	pos *= totLen;
	for (unsigned int i = 0; i < splines.size(); i++) {
		if (pos<lengths[i] || i==splines.size()-1) {
			splines[i]->calc(pos/lengths[i],out);
			return;
		} else {
			pos -= lengths[i];
		}
	}
}

float CompSpline::getLength() {
	return totLen;
}

void rotatePoint(float x, float y, float middleX, float middleY, int rotation, float *outX, float *outY) {
	float relx = x-middleX;
	float rely = y-middleY;
	float angle = rotation*2*M_PI/6;
	*outX = middleX + relx*cos(angle) - rely*sin(angle);
	*outY = middleY + rely*cos(angle) + relx*sin(angle);
}

CompSpline *getSpline(int tile, int rotation, int nr, bool opposite) {
	TileSplines &ts = tileSplines[tile];
	if (nr>=ts.usedCoords) return NULL;
	SplineCoords &sc = ts.coords[nr];
	SplinePoint points[3];
	for (int i = 0; i < sc.usedPoints; i++) {
		int j = i;
		if (opposite) j = sc.usedPoints-1-i;
		rotatePoint(sc.points[i].x,sc.points[i].y,tileWidth/2,tileHeight/2,rotation,&points[j].x,&points[j].y);
		rotatePoint(sc.points[i].relx,sc.points[i].rely,0,0,rotation,&points[j].relx,&points[j].rely);
		if (opposite) {
			points[j].relx = -points[j].relx;
			points[j].rely = -points[j].rely;
		}
	}
	CompSpline *out = new CompSpline();
	for (int i = 0; i < sc.usedPoints; i++)
		out->pushPoint(points[i].x,points[i].y,points[i].relx,points[i].rely);
	return out;
}


